freeze;

/////////////////////////////////////////////////////////////////////////
// boxcar.m
/////////////////////////////////////////////////////////////////////////
// $Revision: 37674 $
// $Date: 2012-03-04 12:20:09 +0000 (Sun, 04 Mar 2012) $
// $LastChangedBy: kasprzyk $
/////////////////////////////////////////////////////////////////////////
// Authors: Alexander Kasprzyk and Dan Roozemond
/////////////////////////////////////////////////////////////////////////
// Provides support for Boxcar notifications via the Growl service.
// For further details about Boxcar, visit the Boxcar home page at:
//  http://boxcar.io/
// The e-mail address must already be associated with a Boxcar account,
// and you must register it with the provider before you first use
// these intrinsics.
// To register your e-mail address, run the boxcar_register script
// located in the same directory as this file.
/////////////////////////////////////////////////////////////////////////

import "utilities.m": escape_string, is_string_integer, default_curl_path;

/////////////////////////////////////////////////////////////////////////
// Constants (feel free to edit these)
/////////////////////////////////////////////////////////////////////////

// The API key -- we use the generic "Monitoring" provider
API_KEY:="MH0S7xOFSwVLNvNhTpiC";

// The name to send the notification from
API_NAME:="Magma";

// The icon Boxcar uses -- available icons are:
//   magma_blue_57.png
//   magma_green_57.png
//   magma_white_57.png
//   magma_yellow_57.png
ICON_URL:="http://magma.maths.usyd.edu.au/images/icons/magma_white_57.png";

/////////////////////////////////////////////////////////////////////////
// Data Store
/////////////////////////////////////////////////////////////////////////

// The local store -- do not access directly!
boxcar_store:=NewStore();

// Stores the e-mail address in the cache.
procedure save_email(email)
    StoreSet(boxcar_store,"email",email);
end procedure;

// Fetches the e-mail address from the cache. Returns true followed by the
// e-mail address on success, false otherwise.
function fetch_email()
    bool,email:=StoreIsDefined(boxcar_store,"email");
    if bool then
        return true,email;
    else
        return false,_;
    end if;
end function;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// Returns the Boxcar command.
function build_boxcar_command(email,message)
    // Create the curl command
    cmd:=default_curl_path() cat " -i -sS --stderr /dev/null -X POST " cat
                        "--data \"email=" cat escape_string(email) cat "\"";
	for i in [<"from_screen_name", API_NAME>,
              <"icon_url", ICON_URL>,
	          <"message", message>] do
        cmd cat:= " --data \"notification[" cat i[1] cat "]=" cat
                           escape_string(i[2]) cat "\"";
	end for;
	cmd cat:= " https://boxcar.io/devices/providers/" cat API_KEY
	                                                  cat "/notifications";
    return cmd;
end function;

// Extracts the status code from the HTML headers. If no "Status: ..." line can
// be found, or if it isn't an integer value, then returns -1.
function extract_html_header_status(headers)
    idx:=Index(headers,"Status:");
    if idx eq 0 then return -1; end if;
    _,val,idx2:=is_string_integer(headers,idx + 8);
    return idx + 8 ne idx2 select val else -1;
end function;

/////////////////////////////////////////////////////////////////////////
// Intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic BoxcarInit( email::MonStgElt )
{Set the e-mail address associated with your Boxcar account}
    // Sanity checks
	require not InternalIsPC(): "Operation not supported on Windows";
    // Check if curl responds
	try
		i:=Pipe(default_curl_path() cat " -s --head http://www.google.com/","");
	catch e
        Trim(e`Object);
		error "Sending Boxcar notifications from Magma requires curl, and curl doesn't seem to work";
	end try;
	// Save the data
	save_email(email);
end intrinsic;

intrinsic Boxcar( message::MonStgElt ) -> BoolElt,MonStgElt
{Send a Boxcar notification. Returns a boolean indicating whether this was a success. If there was a problem, the second returned value will contain a brief error message.}
    // Sanity checks
    bool,email:=fetch_email();
    require bool: "Call BoxcarInit before using Boxcar for the first time";
	// Create the command
	cmd:=build_boxcar_command(email,message);
	// Execute command
	try resp:=Pipe(cmd, "");
	   catch err return false,err;
    end try;
	// Parse the header for the status
	status:=extract_html_header_status(resp);
	// Return success or any errors
	if status eq 200 then
        return true,_;
    end if;
    if status eq 400 then
        return false,"Incorrect parameters.";
    end if;
	if status eq 401 or status eq 403 then
	   return false,"Request failed.";
    end if;
	if status eq 404 then
        return false,"E-mail address is not registered with this provider.";
    end if;
    return false,"Invalid status code [" cat IntegerToString(status) cat "].";
end intrinsic;
