freeze;

/////////////////////////////////////////////////////////////////////////
// growl.m
/////////////////////////////////////////////////////////////////////////
// $Revision: 34307 $
// $Date: 2011-07-09 15:51:17 +0100 (Sat, 09 Jul 2011) $
// $LastChangedBy: kasprzyk $
/////////////////////////////////////////////////////////////////////////
// Authors: Alexander Kasprzyk and Dan Roozemond
/////////////////////////////////////////////////////////////////////////
// Provides support for Growl notifications.
// For further details about Growl, visit the Growl home page at:
//  http://growl.info/
/////////////////////////////////////////////////////////////////////////
// If you only want to recieve notifications originating from your
// own computer (like regular Growl), you should install the
// "growlnotify" script bundled in the "Extras" directory with Growl.
/////////////////////////////////////////////////////////////////////////
// If you want to recieve notifications from machines other than
// your own computer, you will need to enable network notifications
// and registration. The settings are in System Preferences... -> Growl,
// under the "Network" tab.
// If you're running a firewall, your machine will need to be able
// to recieve data over TCP port 23052 and UDP port 9887.
// You can specify the address and password via the environment
// variables:
//  - MAGMAGROWL_ADDRESS 	e.g. "127.0.0.1"
//  - MAGMAGROWL_PASSWORD	e.g. "secret"
/////////////////////////////////////////////////////////////////////////

import "utilities.m": escape_string, is_localhost;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// Returns the default address.
function default_growl_address()
    address:=GetEnv("MAGMAGROWL_ADDRESS");
    return #address eq 0 select "127.0.0.1" else address;
end function;

// Returns the default password.
function default_growl_password()
    return GetEnv("MAGMAGROWL_PASSWORD");
end function;

// Returns true followed by the the path to growlnotify, if available, false.
// otherwise.
function has_growlnotify()
    path:=Pipe("which growlnotify 2>&1 || echo \"FAILED\"","");
    if Index(path,"FAILED") ne 0 then
        return false,_;
    else
        return true,Trim(path);
    end if;
end function;

// Returns the path to the directory containing this file.
function magmagrowl_dir()
    // Magma doesn't provide a way for a package file to discover its
    // path directly -- instead we force an error and parse the error string
    // to extract the path.
    try
        a:=0;
        a:=1/a;
    catch e;
        idx:=Index(e`Position,"\"");
        path:=e`Position[idx+1..#e`Position];
        idx:=Index(path,"\"");
        path:=path[1..idx-1];
        path:=Pipe("dirname \"" cat path cat "\"","");
        idx:=Index(path,"\n");
        if idx ne 0 then
            path:=path[1..idx-1];
        end if;
    end try;
    return path cat "/";
end function;

// Returns the Growl command.
function build_growl_command(title,message,address,password)
    // Get the path to the directory containing this file
    dir:=magmagrowl_dir();
    // Do we have growlnotify?
    bool,path:=has_growlnotify();
    if bool then
        // Is the target the local machine? If so we can use a simplified syntax
        if is_localhost(address) then
            return "echo \"" cat message cat "\" | " cat path cat
                    " --image \"" cat dir cat "magma_white_32.png\"" cat
                    " -n Magma -d Status -t \"" cat title cat "\" 2>&1";
        else
            return "echo \"" cat message cat "\" | " cat path cat
                    " --image \"" cat dir cat "magma_white_32.png\"" cat
                    " -H \"" cat address cat "\" -P \"" cat password cat
                    "\" -n Magma -d Status -t \"" cat title cat "\" 2>&1";
        end if;
    end if;
    // No -- we try the alternative magmagrowl scripts
    args:="\"" cat address cat "\" \"" cat password cat "\" \"" cat title cat
          "\" \"" cat message cat "\"";
    return dir cat "magmagrowl.py " cat args cat "  2>&1 || " cat
           dir cat "magmagrowl.php " cat args cat " 2>&1";
end function;

/////////////////////////////////////////////////////////////////////////
// Intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Growl( title::MonStgElt, message::MonStgElt :
    address:=false, password:=false )
{Send a Growl notification}
    // Sanity checks
    require not InternalIsPC(): "Operation not supported on Windows";
    if address cmpeq false then address:=default_growl_address(); end if;
    if password cmpeq false then password:=default_growl_password(); end if;
    require Type(address) eq MonStgElt: "'address' must be a string";
    require Type(password) eq MonStgElt: "'password' must be a string";
    // Create the command
    title:=escape_string(title);
    message:=escape_string(message);
    cmd:=build_growl_command(title,message,address,password);
    // Open a pipe and send the message
    res:=Pipe(cmd cat " || echo \"FAILED\"","");
    // Was this a success?
    idx:=Index(res,"FAILED");
    if idx ne 0 then
        "Warning: Unable to send Growl notification.";
        if idx gt 1 then Trim(res[1..idx - 1]); end if;
    end if;
end intrinsic;
