freeze;

/////////////////////////////////////////////////////////////////////////
// howl.m
/////////////////////////////////////////////////////////////////////////
// $Revision: 37571 $
// $Date: 2012-02-27 21:21:21 +0000 (Mon, 27 Feb 2012) $
// $LastChangedBy: kasprzyk $
/////////////////////////////////////////////////////////////////////////
// Authors: Alexander Kasprzyk and Dan Roozemond
/////////////////////////////////////////////////////////////////////////
// Provides support for Howl notifications.
// For further details about Howl, visit the Howl home page at:
//  http://howlapp.com/
/////////////////////////////////////////////////////////////////////////
// See http://howlapp.com/interface.html for some documentation. 
// See http://howlapp.com/interface-icons.html for a list of
// acceptable icons.
/////////////////////////////////////////////////////////////////////////

import "utilities.m": escape_string, host_name, default_curl_path;
import "xml.m": XML_parse_string, XML_node_name, XML_get_nodes_by_name, XML_node_children, XML_is_text_node, XML_node_value;

/////////////////////////////////////////////////////////////////////////
// Data Store
/////////////////////////////////////////////////////////////////////////

// The local store -- do not access directly!
howl_store:=NewStore();

// Stores the given data in the cache.
procedure save_data(data)
    StoreSet(howl_store,"data",data);
end procedure;

// Fetches the data from the cache. Returns true followed by the data on
// success, false otherwise.
function fetch_data()
    bool,data:=StoreIsDefined(howl_store,"data");
    if bool then
        return true,data;
    else
        return false,_;
    end if;
end function;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// Returns the Howl command.
function build_howl_command(username,password,
                            name,title,message,application,hostname,icon)
    cmd:=default_curl_path() cat " -sS --stderr /dev/null -X POST";
	for i in [<"name", name>,
	          <"title", title>,
	          <"description", message>,
	          <"application", application>,
	          <"hostname", hostname>,
	          <"icon-name", icon>] do
        cmd cat:= " --data \"" cat i[1] cat "=" cat
                           escape_string(i[2]) cat "\"";
	end for;
	cmd cat:= " -u \"" cat escape_string(username) cat ":" cat
                           escape_string(password) cat "\"" cat
	                       " https://howlapp.com/public/api/notification.xml";
    return cmd;
end function;

/////////////////////////////////////////////////////////////////////////
// Intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic HowlInit( username::MonStgElt, password::MonStgElt )
{Set the username and password for your Howl account}
    // Sanity checks
	require not InternalIsPC(): "Operation not supported on Windows";
    if Position(password, ":") ne 0 then
        "WARNING: Your Howl password contains a colon -- no idea what will happen on the command line!";
	end if;
    // Check if curl responds
	try
		i:=Pipe(default_curl_path() cat " -s --head http://www.google.com/","");
	catch e
        Trim(e`Object);
		error "Howling from Magma requires curl, and curl doesn't seem to work";
	end try;
	// Save the values
	save_data([username,password]);
end intrinsic;

intrinsic Howl( title::MonStgElt, message::MonStgElt :
    application:="Magma", name:="howl", icon:="warning", hostname:=false )
    -> BoolElt,MonStgElt
{Send a Howl notification. Returns a boolean indicating whether this was a success. If there was a problem, the second returned value will contain a brief error message.}
    // Sanity checks
    bool,data:=fetch_data();
    require bool: "Call HowlInit before using Howl for the first time";
    require Type(application) eq MonStgElt: "'application' must be a string";
    require Type(name) eq MonStgElt: "'name' must be a string";
    require Type(icon) eq MonStgElt: "'icon' must be a string";
    if hostname cmpeq false then
        hostname:=host_name();
    end if;
    require Type(hostname) eq MonStgElt: "'hostname' must be a string";
	// Extract the username and password
	username:=data[1];
	password:=data[2];
	// Create the command
	cmd:=build_howl_command(username,password,
	                        name,title,message,application,hostname,icon);
	// Execute command
	try resp:=Pipe(cmd, "");
	   catch err return false,err;
    end try;
	// Parse result as XML
	try xml:=XML_parse_string(resp);
	   catch err return false,"Could not parse XML response";
    end try;
	// Analyse XML
	try 
		xml:=XML_parse_string(resp); 
		assert XML_node_name(xml) eq "plist";
		dict:=XML_get_nodes_by_name(xml, "dict");
		assert #dict eq 1;
		dict:=dict[1];
		status:=false;
		children:=XML_node_children(dict);
		for i in [1..#children-1] do
			node:=children[i];
		    name:=XML_node_name(node);
		    if name eq "key" then
		        assert XML_is_text_node(node);
				vl:=XML_node_value(node);
				if vl eq "status" then
					nextnode:=children[i+1];
					assert XML_is_text_node(nextnode);
					status:=XML_node_value(nextnode);
					break i;
		        end if;
		    end if;
		end for;
		assert status cmpne false;
	catch err 
		return false,"Could not parse XML response"; 
	end try;
	// Check and return the status
	if status cmpeq "ok" then
	   return true,_;
    else
        return false,status;
    end if;
end intrinsic;
